from .command import CmdOpt, CommandMaster
from .config import APP_ENV, Config, Env
from .convert import MConvert
from .dict import MDict
from .fetch import MFetch
from .file import MFile
from .funny import advance_call, advance_class_call, async_advance_call
from .list import MList
from .retry import async_retry, retry
from .url import MUrl



import sys
from enum import Enum
from pathlib import Path
from typing import List, Optional, Union

import yaml
from pydantic import BaseModel

_config_root = Path("./configs")


class Env(Enum):
    """预设环境变量"""

    Local = "local"
    Dev = "dev"
    Test = "test"
    Prod = "prod"
    Custom = "custom"
    Default = "dev"  # noqa: PIE796


APP_ENV: str = Env.Default.value


class Config(BaseModel):
    """配置基类"""

    @classmethod
    def set_configs_root(cls, root: Union[str, Path]):
        """设置配置文件根目录"""
        global _config_root
        _config_root = Path(root)

    @classmethod
    def load_config(cls, create_if_not_exists: bool = True):
        """加载配置

        :param create_if_not_exists: 如果配置文件不存在，是否创建

        :raises Exception: 配置文件格式错误
        """

        global APP_ENV, _config_root
        config_path = _config_root / f"config.{APP_ENV}.yaml"
        config_path.parent.mkdir(parents=True, exist_ok=True)
        try:
            with config_path.open("r") as file:
                # obj = cls.parse_obj(yaml.safe_load(file))
                obj = cls.model_validate(yaml.safe_load(file))
        except Exception:
            if create_if_not_exists:
                cls.dump_config_template()
                return cls.load_config(False)
            raise
        try:
            return cls.model_validate(obj)
        except Exception as e:
            raise Exception(f"配置文件格式错误: {e}") from e

    @classmethod
    def dump_modal(cls):
        """导出配置模板

        :return: 配置模板
        """
        return cls().__dict__

    @classmethod
    def dump_config_template(cls, envs: Optional[List[str]] = None):
        """导出配置模板"""

        global _config_root
        if envs is None:
            envs = [APP_ENV]
        for env in envs:
            config_path = _config_root / f"config.{env}.yaml"
            config_path.parent.mkdir(parents=True, exist_ok=True)
            config_path.write_text(
                yaml.dump(cls.dump_modal(), allow_unicode=True, sort_keys=False),
            )
        return cls.dump_modal()

    def dump_config(self, envs: Optional[List[str]] = None):
        """导出配置"""

        global _config_root
        if envs is None:
            envs = [APP_ENV]
        for env in envs:
            config_path = _config_root / f"config.{env}.yaml"
            config_path.parent.mkdir(parents=True, exist_ok=True)
            if config_path.exists():
                # 如果配置文件存在，则只补充缺失的配置项
                origin_config = yaml.safe_load(config_path.read_text())
                for key, value in self.__dict__.items():
                    if key not in origin_config:
                        origin_config[key] = value
                config_path.write_text(
                    yaml.dump(origin_config, allow_unicode=True, sort_keys=False),
                )
            else:
                config_path.write_text(
                    yaml.dump(self.model_dump(), allow_unicode=True, sort_keys=False),
                )

    def reload_config(self):
        """重新加载配置"""
        obj = self.load_config()
        self.__dict__.update(obj.__dict__)

    def gen_config_schema(self):
        """生成配置结构

        :return: 配置结构
        """
        return self.__class__.model_json_schema()["properties"]


for arg in sys.argv[1:]:
    if arg.startswith("env="):
        target_env = arg.split("=")[-1]
        for env in Env:
            if env.value.lower() == target_env.lower():
                APP_ENV = env.value
                break
        else:
            raise Exception(f'环境变量 "{target_env}" 不存在')
        break
else:
    APP_ENV = Env.Default.value



import asyncio
import time


def retry(retry_times=3, delay=1, ignore_exception: bool = True):
    """重试装饰器"""

    def wrapper(func):
        def inner(*args, **kwargs):
            for i in range(retry_times):
                try:
                    return func(*args, **kwargs)
                except Exception:
                    if i == retry_times - 1:
                        if ignore_exception:
                            pass
                        else:
                            raise
                    else:
                        time.sleep(delay)
            return None

        return inner

    return wrapper


def async_retry(retry_times=3, delay=1, ignore_exception: bool = True):
    """异步重试装饰器"""

    def wrapper(func):
        async def inner(*args, **kwargs):
            for i in range(retry_times):
                try:
                    return await func(*args, **kwargs)
                except Exception:
                    if i == retry_times - 1:
                        if ignore_exception:
                            pass
                        else:
                            raise
                    else:
                        await asyncio.sleep(delay)
            return None

        return inner

    return wrapper



from typing import List

try:
    import ujson as json
except ImportError:
    import json


class MList(list):
    """列表类"""

    @classmethod
    def quick_value(cls, ls: List, substitute: str = "") -> str:
        """快速列表取值"""
        if isinstance(ls, str):
            return ls.strip() if (ls and ls.strip()) else substitute
        if len(ls) == 0:
            return substitute
        return ls[0].strip() if (ls[0] and ls[0].strip()) else substitute

    @classmethod
    def drop_empty_from_list(cls, lst: List) -> List[str]:
        """去除列表中的空字符串元素"""

        return [i for i in lst if i and i.strip()]

    @classmethod
    def split_and_drop_empty(cls, s: str, sep: str) -> List[str]:
        """切割字符串并去除空字符串元素"""
        return cls.drop_empty_from_list(s.split(sep))

    @classmethod
    def json_list_stringify_limit(cls, lst: List, limit=1024) -> str:
        """列表转json列表长度自动限制(丢弃超出限制的元素)"""
        while True:
            if len(lst) < 1:
                return "[]"
            res = json.dumps(lst, ensure_ascii=False)
            if len(res) <= limit:
                return res
            lst = lst[:-1]

    @classmethod
    def advance_split(cls, s: str, sep: str, filter_empty=True) -> List[str]:
        """高级切割字符串 (支持单双引号配对防截断和引号转义)"""
        res = []
        temp = ""
        in_single_quote = False
        in_double_quote = False
        escape = False
        for i in s:
            if escape:
                temp += i
                escape = False
                continue
            if i == "\\":
                escape = True
                continue
            if i == "'":
                in_single_quote = not in_single_quote
            if i == '"':
                in_double_quote = not in_double_quote
            if i == sep and not in_single_quote and not in_double_quote:
                res.append(temp)
                temp = ""
            else:
                temp += i
        res.append(temp)

        # 过滤空字符串
        if filter_empty:
            res = [i for i in res if i.strip()]

        # 去除引号
        res = [i[1:-1] if i.startswith('"') and i.endswith('"') else i for i in res]
        res = [i[1:-1] if i.startswith("'") and i.endswith("'") else i for i in res]

        return res  # noqa: RET504



import asyncio
import time

SPINNER = ["|", "/", "-", "\\"]


def advance_call(
    show_output=True,
    spin_interval=0.1,
    loops=1,
    output_format=None,
    failed_callback=None,
):
    """\"高级\"调用装饰器

    Args:
        show_output (bool, optional): 是否显示输出. Defaults to True.
        spin_interval (float, optional): 旋转间隔. Defaults to 0.1.
        loops (int, optional): 旋转次数. Defaults to 1.
        output_format (str, optional): 输出格式. Defaults to None. 要自定义输出格式, 请使用{char}作为旋转字符占位符.
    """

    def wrapper(func):
        def inner(*args, **kwargs):
            if show_output:
                for _ in range(loops):
                    for char in SPINNER:
                        if output_format:
                            print(output_format.format(char), end="\r")
                        else:
                            print(f"[{char}] Calling {func.__name__}...", end="\r")
                        time.sleep(spin_interval)

            try:
                func(*args, **kwargs)
            except Exception as e:
                if failed_callback:
                    failed_callback(e)
                else:
                    raise

        return inner

    return wrapper


def async_advance_call(
    show_output=True,
    spin_interval=0.1,
    loops=1,
    output_format=None,
):
    """\"高级\"异步调用装饰器

    Args:
        show_output (bool, optional): 是否显示输出. Defaults to True.
        spin_interval (float, optional): 旋转间隔. Defaults to 0.1.
        loops (int, optional): 旋转次数. Defaults to 1.
    """

    def wrapper(func):
        async def inner(*args, **kwargs):
            if show_output:
                for _ in range(loops):
                    for char in SPINNER:
                        if output_format:
                            print(output_format.format(char), end="\r")
                        else:
                            print(f"[{char}] Calling {func.__name__}...", end="\r")
                        await asyncio.sleep(spin_interval)

            await func(*args, **kwargs)

        return inner

    return wrapper


def advance_class_call(
    show_output=True,
    spin_interval=0.1,
    loops=1,
    output_format=None,
):
    """\"高级\"类调用装饰器

    Args:
        show_output (bool, optional): 是否显示输出. Defaults to True.
        spin_interval (float, optional): 旋转间隔. Defaults to 0.1.
        loops (int, optional): 旋转次数. Defaults to 1.
    """

    def wrapper(cls):
        class Inner(cls):
            def __init__(self, *args, **kwargs):
                if show_output:
                    for _ in range(loops):
                        for char in SPINNER:
                            if output_format:
                                print(output_format.format(char), end="\r")
                            else:
                                print(
                                    f"[{char}] Initializing {cls.__name__}...",
                                    end="\r",
                                )
                            time.sleep(spin_interval)

                super().__init__(*args, **kwargs)

        return Inner

    return wrapper


def with_pray(show_output=True, spin_interval=0.3, loops=3):
    def on_failed(_):
        print("The Coding God is very angry with you.")

    return advance_call(
        show_output=show_output,
        spin_interval=spin_interval,
        loops=loops,
        output_format="[{}] Praying to the Coding God for no error...",
        failed_callback=on_failed,
    )
